//*******************************************************************
//* A Custon BLE Design Using the Raytac MDBT50Q Module
//* REV 1.0.0
//* LAST UPDATE 04-05-2024
//* NOTES:
//*******************************************************************

#include <stdbool.h>
#include <zephyr/types.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/kernel.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>

#include "cecsvc.h"

#include <zephyr/logging/log.h>

LOG_MODULE_DECLARE(bleday4, LOG_LEVEL_DBG);

static struct sensorsvc_cb		sensor_cb;
static struct ledsvc_cb 		led_cb;

static uint16_t adc_val;
static bool status_notify;

static void status_cccd_changed(const struct bt_gatt_attr *attr,uint16_t value)
{
	LOG_DBG("status_cccd_changed called");
	status_notify = value == BT_GATT_CCC_NOTIFY;
}


//*******************************************************************
//* Read Temp Sensor Callback
//*******************************************************************
static ssize_t read_temp_sensor(struct bt_conn *conn,
			  const struct bt_gatt_attr *attr,
			  void *buf,
			  uint16_t len,
			  uint16_t offset)
{
	//get a pointer to sensor value which is passed in the BT_GATT_CHARACTERISTIC() and stored in attr->user_data
	const uint16_t *value = attr->user_data;
	LOG_INF("Attribute read, handle: %u, conn: %p", attr->handle, (void *)conn);
	if (sensor_cb.sensorchrc_cb) {
		// Call the application callback function to update the get the current adc value
		adc_val = sensor_cb.sensorchrc_cb();

		LOG_INF("adc_val: %04x",adc_val);
		LOG_INF("len: %u",len);
		LOG_INF("offset: %u",offset);
		LOG_INF("size of value: %u",sizeof(*value));
		LOG_INF("value: %d",*value);
		return bt_gatt_attr_read(conn, attr, buf, len, offset, value, sizeof(*value));
	}
	return 0;
}

//*******************************************************************
//* Create and add the SENSOR service to the Bluetooth LE stack 
//*******************************************************************
BT_GATT_SERVICE_DEFINE(sensor_svc,
BT_GATT_PRIMARY_SERVICE(BT_UUID_SENSORSVC),
// Create and add the sensor characteristic 
BT_GATT_CHARACTERISTIC(BT_UUID_SENSORCHR,
			BT_GATT_CHRC_READ | BT_GATT_CHRC_NOTIFY, 
			BT_GATT_PERM_READ, read_temp_sensor, NULL,
			&adc_val),
BT_GATT_CCC(status_cccd_changed,
			BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),

);

void status_nfy(struct bt_conn *conn, uint16_t adcValue)
{
	uint8_t buf[2];
	buf[0] = (adcValue & 0xFF00) >> 8;
	buf[1] = (uint8_t) adcValue & 0x00FF;

	bt_gatt_notify(conn, &sensor_svc.attrs[1],buf,sizeof(buf));
}

//*******************************************************************
//* REGISTER APPLICATION CALLBACKS
//*******************************************************************
// A function to register application callbacks for the sensor characteristic
int sensorsvc_init(struct sensorsvc_cb *sensor_callback)
{
	if (sensor_callback) {
		sensor_cb.sensorchrc_cb = sensor_callback->sensorchrc_cb;
	}
	return 0;
}

//*******************************************************************
//* LED Control Callback
//*******************************************************************
static ssize_t write_led(struct bt_conn *conn,
			 const struct bt_gatt_attr *attr,
			 const void *buf,
			 uint16_t len, uint16_t offset, uint8_t flags)
{
	LOG_INF("Attribute write, handle: %u, conn: %p", attr->handle,
		(void *)conn);
	if (len != 1U) {
		LOG_INF("Incorrect data length");
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_ATTRIBUTE_LEN);
	}
	if (offset != 0) {
		LOG_INF("Incorrect data offset");
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
	}
	if (led_cb.ledchrc_cb) {
		//Read the received value 
		uint8_t val = *((uint8_t *)buf);
		if (val == 0x00 || val == 0x01) {
			//Call the application callback function to update the led state
			led_cb.ledchrc_cb(val ? 0x01 : 0x00);
		} 
		else {
			LOG_INF("Incorrect value");
			return BT_GATT_ERR(BT_ATT_ERR_VALUE_NOT_ALLOWED);
		}
	}
	return len;
}

//*******************************************************************
//* Create and add the LED control service to the Bluetooth LE stack 
//*******************************************************************
BT_GATT_SERVICE_DEFINE(led_svc,
BT_GATT_PRIMARY_SERVICE(BT_UUID_LEDSVC),
/* Create and add the LED characteristic */
		       BT_GATT_CHARACTERISTIC(BT_UUID_LEDCHR,
			   			BT_GATT_CHRC_WRITE,
						BT_GATT_PERM_WRITE,
						NULL, write_led, NULL),
);

//*******************************************************************
//* REGISTER APPLICATION CALLBACKS
//*******************************************************************
// A function to register application callbacks for the led characteristic
int ledsvc_init(struct ledsvc_cb *led_callback)
{
	if (led_callback) {
		led_cb.ledchrc_cb = led_callback->ledchrc_cb;
	}
	return 0;
}

