//*******************************************************************
//* A Custon BLE Design Using the Raytac MDBT50Q Module
//* REV 1.0.0
//* LAST UPDATE 04-05-2024
//* NOTES:
//*******************************************************************

#include <inttypes.h>
#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>
#include <zephyr/types.h>
#include <zephyr/kernel.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <soc.h>

#include <zephyr/logging/log.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/gap.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/addr.h>

#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/adc.h>
#include <zephyr/sys/util.h>
#include <zephyr/sys/printk.h>

#include "cecsvc.h"

void status_nfy(struct bt_conn *conn, uint16_t adcValue);

int err;
int16_t bufadc;
int val_mv;
uint8_t flags;
uint16_t adc0_val;
uint16_t old_adc0_val;
#define fconnected		0b00000001
#define SET_FLAG(flagbit) (flags |= flagbit)
#define CLR_FLAG(flagbit) (flags &= ~flagbit)
#define CHK_FLAG(flagbit) (flags & flagbit)
#define TOG_FLAG(flagbit) (flags ^= flagbit)

LOG_MODULE_REGISTER(bleday4, LOG_LEVEL_DBG);

struct adc_sequence sequence = {
	.buffer = &bufadc,
	// buffer size in bytes, not number of samples 
	.buffer_size = sizeof(bufadc),
};

#if !DT_NODE_EXISTS(DT_PATH(zephyr_user)) || \
	!DT_NODE_HAS_PROP(DT_PATH(zephyr_user), io_channels)
#error "No suitable devicetree overlay specified"
#endif

#define DT_SPEC_AND_COMMA(node_id, prop, idx) \
	ADC_DT_SPEC_GET_BY_IDX(node_id, idx),

// Data of ADC io-channels specified in devicetree.
static const struct adc_dt_spec adc_channels[] = {
	DT_FOREACH_PROP_ELEM(DT_PATH(zephyr_user), io_channels, DT_SPEC_AND_COMMA)
};

#define RUN_STATUS_LED 	DT_ALIAS(led1)
#define CON_STATUS_LED 	DT_ALIAS(led2)
#define LED_USER 		DT_ALIAS(led3)
#define LED_CNTL 		DT_ALIAS(led4)

static const struct gpio_dt_spec runstatus = GPIO_DT_SPEC_GET(RUN_STATUS_LED,gpios);
static const struct gpio_dt_spec constatus = GPIO_DT_SPEC_GET(CON_STATUS_LED,gpios);
static const struct gpio_dt_spec leduser = GPIO_DT_SPEC_GET(LED_USER,gpios);
static const struct gpio_dt_spec ledcntl = GPIO_DT_SPEC_GET(LED_CNTL,gpios);

#define DEVICE_NAME CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN (sizeof(DEVICE_NAME) - 1)

#define RUN_LED_BLINK_INTERVAL 1000

static struct bt_le_adv_param *adv_param = BT_LE_ADV_PARAM(
	(BT_LE_ADV_OPT_CONNECTABLE |
	 BT_LE_ADV_OPT_USE_IDENTITY), // Connectable advertising and use identity address
	800, // Min Advertising Interval 500ms (800*0.625ms)
	801, // Max Advertising Interval 500.625ms (801*0.625ms)
	NULL); // Set to NULL for undirected advertising

static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
};

static const struct bt_data sd[] = {
	BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_LEDSVC_VAL),
};

//*******************************************************************
//* Read Temp Sensor Callback
//* called by read_temp_sensor() in cecsvc.c
//* Temp C = (val_mv - 500) / 10
//*******************************************************************
static uint16_t sensor_cb(void)
{
		(void)adc_sequence_init_dt(&adc_channels[0], &sequence);

		err = adc_read(adc_channels[0].dev, &sequence);
		if (err < 0) {
			LOG_ERR("Could not read (%d)\n", err);
		} 
		//LOG_INF("bufadc: 0x%02X",bufadc);
		val_mv = (int)bufadc;
		err = adc_raw_to_millivolts_dt(&adc_channels[0],&val_mv);
		if(err<0){
			LOG_INF("raw to mv no go\n");
		}
		//else {
		//	LOG_INF("%d mV",val_mv - 500);
		//}
		
	return val_mv - 500;
}

//*******************************************************************
//* Declare a variable sensor_callback of type sensorsvc_cb and
//* initiate its members to the applications call back functions
//*******************************************************************

static struct sensorsvc_cb sensor_callback = {
	.sensorchrc_cb = sensor_cb,
};

//*******************************************************************
//* LED Control Callback
//* called by write_led() in cecsvc.c
//*******************************************************************
static void ledchrc_cb(uint8_t enablebite)
{
	switch(enablebite)
	{
		case 0:
			LOG_INF("LED is OFF.\n");
			gpio_pin_set_dt(&ledcntl, 0);
		break;

		case 1:
			LOG_INF("LED is ON.\n");
			gpio_pin_set_dt(&ledcntl, 1);
		break;
	}
}

//*******************************************************************
//* Declare a variable sensor_callbacks of type ledsvc_cb and
//* initiate its members to the applications call back functions
//*******************************************************************
static struct ledsvc_cb led_callback = {
	.ledchrc_cb = ledchrc_cb,
};

//*******************************************************************
//* Connected callback
//*******************************************************************
static void on_connected(struct bt_conn *conn, uint8_t err)
{
	if (err) {
		LOG_ERR("Connection failed (err %u)\n", err);
		return;
	}

	LOG_INF("Connected\n");

	gpio_pin_set_dt(&constatus, 1);
	SET_FLAG(fconnected);

	struct bt_conn_info info;
	err = bt_conn_get_info(conn, &info);
	if (err) {
   		LOG_ERR("bt_conn_get_info() returned %d", err);
   		return;
	}	
}

//*******************************************************************
//* Disconnected callback
//*******************************************************************
static void on_disconnected(struct bt_conn *conn, uint8_t reason)
{
	LOG_INF("Disconnected (reason %u)\n", reason);

	gpio_pin_set_dt(&constatus, 0);
	CLR_FLAG(fconnected);
}

//*******************************************************************
//* Declare a variable connection_callbacks of type bt_conn_cb and
//* initiate its members to the applications call back functions
//*******************************************************************
struct bt_conn_cb connection_callbacks = {
	.connected = on_connected,
	.disconnected = on_disconnected,
};

//*******************************************************************
//* Configure GPIO Function
//*******************************************************************
static void configure_gpio(void)
{
	gpio_pin_configure_dt(&runstatus, GPIO_OUTPUT_INACTIVE);
	gpio_pin_configure_dt(&constatus, GPIO_OUTPUT_INACTIVE);
	gpio_pin_configure_dt(&leduser, GPIO_OUTPUT_INACTIVE);
	gpio_pin_configure_dt(&ledcntl, GPIO_OUTPUT_INACTIVE);
}

//*******************************************************************
//* MAIN Function
//*******************************************************************
int main(void)
{
	int blink_status = 0;
	int err;

	LOG_INF("A Custom BLE Design Using the Raytac MDBT50Q BLE Module\n");

	configure_gpio();

	// Configure channels individually prior to sampling. 
	for (size_t i = 0U; i < ARRAY_SIZE(adc_channels); i++) {
		if (!device_is_ready(adc_channels[i].dev)) {
			LOG_ERR("ADC controller device not ready\n");
			return -1;
		}

		err = adc_channel_setup_dt(&adc_channels[i]);
		if (err < 0) {
			LOG_ERR("Could not setup channel #%d (%d)\n", i, err);
			return -1;
		}
	}

	err = bt_enable(NULL);
	if (err) {
		LOG_ERR("Bluetooth init failed (err %d)\n", err);
		return -1;
	}
	LOG_INF("Bluetooth initialized\n");

	bt_conn_cb_register(&connection_callbacks);

	bt_addr_le_t addr;
    err = bt_addr_le_from_str("FF:EE:DD:CC:BB:AA", "random", &addr);
    if (err) {
        LOG_ERR("Invalid BT address (err %d)\n", err);
    }
 
	// Pass your application callback functions stored in xxxxx_callbacks to the respective services 
	err = sensorsvc_init(&sensor_callback);
	if (err) {
		LOG_ERR("Failed to init adc callback (err:%d)\n", err);
		return -1;
	}

	err = ledsvc_init(&led_callback);
	if (err) {
		LOG_ERR("Failed to init led callback (err:%d)\n", err);
		return -1;
	}

	err = bt_le_adv_start(adv_param, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
	if (err) {
		LOG_ERR("Advertising failed to start (err %d)\n", err);
		return -1;
	}

	LOG_INF("Advertising successfully started\n");

	old_adc0_val = 0x0000;
	adc0_val = 0x0000;

	for (;;) 
	{
		gpio_pin_set_dt(&runstatus, (++blink_status) % 2);
		k_sleep(K_MSEC(RUN_LED_BLINK_INTERVAL));

		if(CHK_FLAG(fconnected))
		{
			adc0_val = sensor_cb();
			LOG_INF("adc0 = %04X",adc0_val);
			LOG_INF("adc0 %d mV",adc0_val);

			if(old_adc0_val != adc0_val)
			{
				old_adc0_val = adc0_val;
				status_nfy(NULL,adc0_val);

			}
		}
	}
}
